from __future__ import absolute_import
from .redis_object import RedisObject
from .redis_list import RedisList
from .redis_dict import RedisDict
from .redis_set import RedisSet
from .redis_sortedset import RedisSortedSet

__all__ = [
    'RedisObject',
    'RedisList',
    'RedisDict',
    'RedisSet',
    'RedisSortedSet'
]


from . import RedisObject

class RedisSet(RedisObject):
    '''
    An equivalent to set where all items are stored in Redis.
    '''
    def __init__(self, id=None, item_type=str, items=None):
        '''
        Create a new RedisSet
        :param id: If specified, use this as the redis ID, otherwise generate a random ID.
        :param item_type: The constructor to use when reading items from redis.
        :param items: Default values to store during construction.
        '''
        super(RedisSet, self).__init__(id)
        self.item_type = item_type

        if items:
            for item in items:
                self.add(item)

    @classmethod
    def as_child(cls, parent, tag, item_type):
        '''
        Alternative callable constructor that instead defines this as a child object
        :param parent: Parent instance.
        :param tag: Tag of the child object
        :param item_type: Type of the object
        '''
        def helper(_ = None):
            return cls(parent.id + ':' + tag, item_type)
        return helper

    def __len__(self):
        '''
        :return: length of the set
        '''
        return self.redis.scard(self.id)

    def __iter__(self):
        '''
        Iterate over all items in this set.
        '''
        for item in self.redis.smembers(self.id):
            yield self.decode_value(type(item), item)

    def add(self, item):
        '''
        Add an item to the set
        :param item:
        :return: None
        '''
        self.redis.sadd(self.id, self.encode_value(item))

    def remove(self, item):
        '''
        Remove an item from the set.
        :param item:
        :return: None
        '''
        self.redis.srem(self.id, item)

    def ismember(self, item):
        '''
        Return a boolean indicating if item is a member of set
        '''
        return self.redis.sismember(self.id, self.encode_value(item))

    def randmember(self, number=1):
        '''
        returns a list of number random memebers of set.
        '''
        return self.redis.srandmember(self.id, number)

    def pop(self, number=1):
        '''
        Remove and return a random member of set
        '''
        return self.redis.spop(self.id, number)

    def members(self):
        '''
        Return all members of the set
        '''
        return self.redis.smembers(self.id)




import base64
import redis
import os
from .redis_settings import redis_config


class RedisObject(object):
    '''
    A base object backed by redis.
    Genrally, use RedisDict, RedisList, RedisSet, or RedisSortedSet rather than
    this class directly.
    '''

    def __init__(self, id=None):
        '''
        Creae a StrictRedis object. Host, port, db, and password are defined thorough a dictionary where
        you can find at redis_settings.config.

        :param id: (optional) If specified, use this as the redis ID, otherwise generate a random ID.
        '''
        self.redis = redis.StrictRedis(
            host=redis_config['host'],
            port=redis_config['port'],
            db=redis_config['db'],
            password=redis_config['password'],
            decode_responses=True)

        self.id = id if id else base64.urlsafe_b64encode(os.urandom(9)).decode('utf-8')
        if ':' not in self.id:
            self.id = self.__class__.__name__ + ':' + self.id

    def __bool__(self):
        '''
        Test if an object currently exists
        '''
        return self.redis.exists(self.id)

    def __eq__(self, other):
        '''
        Tests if two redis objects are equal (they have the same key)
        '''
        try:
            return self.id == other.id
        except AttributeError as e:
            return False

    def __str__(self):
        '''
        Return this object as a string for testing purposes.
        '''
        return self.id

    def delete(self):
        '''
        Delete this object from redis
        '''
        self.redis.delete(self.id)

    @staticmethod
    def decode_value(type, value):
        '''
        Decode a value if it is non-None, otherwise, decode with no arguments.
        '''
        if value == None:
            return type()
        else:
            return type(value)

    @staticmethod
    def encode_value(value):
        '''Encode a value using str'''
        return str(value)


from . import RedisObject

class RedisSortedSet(RedisObject):
    '''
    A wrapper class for Redis's sorted set.
    '''
    def __init__(self, id=None, init_items=None):
        '''
        Create a new RedisSortedSet
        :param id: If specified, use this as the redis ID, otherwise generate a random ID.
        :param init_items: Default value/score sets to store during construnction.
        '''
        super(RedisSortedSet, self).__init__(id)

        if init_items:
            for item in init_items:
                key = list(item.keys())[0]
                self.add(key, item[key])

    @classmethod
    def as_child(cls, parent, tag):
        '''
        Alternative callable constructor that instead defines this as a child object
        :param parent: Parent instance.
        :param tag: Tag of the child object
        '''
        def helper(_ = None):
            return cls(parent.id + ':' + tag)
        return helper

    def __len__(self):
        '''
        :return: length of the set
        '''
        return self.redis.zcard(self.id)

    def add(self, key, score):
        '''
        Set key/score pair to the sorted set
        :param key:
        :param score:
        :return:
        '''
        self.redis.zadd(self.id, score, key)

    def count(self, min, max):
        '''
        Returns the number of elements in the sorted set with a score between
        min and max.
        '''
        return self.redis.zcount(self.id, min, max)

    def incrby(self, value, amount=1):
        '''
        Increment the score of value in sortedset by amount
        '''
        return self.redis.zincrby(self.id, value, amount)

    def range(self, start, end, desc=False):
        '''
        Return a range of values from sorted set between start and end.
        :param desc: If true, returned values is sored in descending order.
        '''
        return self.redis.zrange(self.id, start, end, desc, withscores=True)

    def revrange(self, start, end):
        '''
        Return a range of values from sorted set between start and end sorted
        in descending order.
        '''
        return self.redis.zrevrange(self.id, start, end, withscores=True)

    def rangebyscore(self, min, max, start=None, num=None):
        '''
        Return a range of values from the sorted set with scores between min and
        max. If start and num are specified, then return a slice of the range.
        '''
        return self.redis.zrangebyscore(self.id, min, max, start, num, withscores=True)

    def rank(self, value):
        '''
        Returns a 0-based value indicating the rank of value in sorted set
        '''
        return self.redis.zrank(self.id, value)

    def rem(self, *values):
        '''
        Remove member values from sorted set
        '''
        return self.redis.zrem(self.id, *values)

    def remrangebyrank(self, min, max):
        '''
        Remove all elements in the sorted set with ranks between min and max.
        Values are 0-based, ordered from smallest score to largest.
        Values can be negative indicating the highest scores.
        Returns the number of elements removed
        '''
        return self.redis.zremrangebyrank(self.id, min, max)

    def remrangebyscore(self, min, max):
        '''
        Remove all elements in the sorted set with scores between min and max.
        Returns the number of elements removed.
        '''
        return self.redis.zremrangebyscore(self.id, min, max)

    def revrangebyscore(self, start, end):
        '''
        Return a range of values from the sorted set with scores between
        min and max in descending order.
        If start and num are specified, then return a slice of the range.
        '''
        return self.redis.zrevrangebyscore(self.id, start, end, withscores=True)

    def revrank(self, value):
        '''
        Returns a 0-based value indicating the descending rank of value in sorted set
        '''
        return self.redis.zrevrank(self.id, value)

    def score(self, value):
        '''
        Return the score of element value in sorted set
        '''
        return self.redis.zscore(self.id, value)



from os.path import join, dirname
from dotenv import load_dotenv
import os

dotenv_path = join(dirname(__file__), '.env')
load_dotenv(dotenv_path)

'''
Configuration for Redis. 

Although this package is using python-dotenv, it is advised to configure
through environment valuables. 
'''

redis_config = {
    "host": os.getenv("REDIS_HOST", "127.0.0.1"),
    "port": int(os.getenv("REDIS_PORT", "6379")),
    "db": int(os.getenv("REDIS_DB", "0")),
    "password": os.getenv("REDIS_PASSWORD", None)
}


from . import RedisObject

class RedisDict(RedisObject):
    '''
    An equivalent to dict where all keys/values are stored in Redis.
    '''
    def __init__(self, id=None, fields={}, defaults=None):
        '''
        Create a RedisDict.
        :param id: If specified, use this as the redis ID, otherwise generate a random ID.
        :param fields: A map of field name to construtor used to read values from redis.
            Objects will be written with json.dumps with default = str, so override __str__ for custom objects.
            This should generally be set by the subobject's constructor.
        :param defaults: Initial value for each field.
        '''
        super(RedisDict, self).__init__(id)
        self.fields = fields
        if defaults:
            for key, val in defaults.items():
                self[key] = val

    def __getitem__(self, key):
        '''
        Load a field from this redis object.
        :param key: The key of the field
        :return: The value of the field. If key doesn't exist, KeyError is raised.
        '''
        if key == 'id':
            return self.id.rsplit(':', 1)[-1]

        if not key in self.fields:
            raise KeyError('{} not found in {}'.format(key, self))

        return RedisObject.decode_value(self.fields[key], self.redis.hget(self.id, key))

    def __setitem__(self, key, val):
        '''
        Store a value in this redis object.
        :param key: The key of the field. If key doesn't exist, KeyError is raised.
        :param val: The value of the field.
        :return: 1 if new field is created, 0 otherwise.
        '''
        if not key in self.fields:
            raise KeyError('{} not found in {}'.format(key, self))
        self.redis.hset(self.id, key, RedisObject.encode_value(val))

    def __iter__(self):
        '''
        :return:  (key, val) pairs for all values stored in this RedisDict.
        '''
        yield ('id', self.id.rsplit(':', 1)[-1])
        for key in self.fields:
            yield (key, self[key])

    def __len__(self):
        '''
        :return: Number of fields of the object.
        '''
        return self.redis.hlen(self.id)

    def __delitem__(self, key):
        '''
        Delete the object.
        :param key: Key of the object.
        :return: None
        '''
        self.redis.hdel(self.id, key)

    def incrby(self, key, amount=1):
        '''
        Increment the field.
        :param key: Key of the field.
        :param amount: Incremental amount.
        :return: incremented value
        '''
        return self.redis.hincrby(self.id, key, amount)

    def incrbyfloat(self, key, amount=1.0):
        '''
        Increment the field.
        :param key: Key of the field.
        :param amount: Incremental amount.
        :return: incremented value
        '''
        return self.redis.hincrbyfloat(self.id, key, amount)

    def keys(self):
        '''
        Get keys of the object.
        :return: Keys of the object
        '''
        return self.redis.hkeys(self.id)

    def values(self):
        '''
        Get values of the object
        :return:  Values of the object
        '''
        return self.redis.hvals(self.id)



from . import RedisObject

class RedisList(RedisObject):
    '''
    An equivalent to list where all items are stored in Redis.
    '''
    def __init__(self, id=None, item_type=str, items=None):
        '''
        Create a new RedisList
        :param id: If specified, use this as the redis ID, otherwise generate a random ID.
        :param item_type: The constructor to use when reading items from redis.
        :param items: Default values to store during construction.
        '''
        super(RedisList, self).__init__(id)
        self.item_type = item_type

        if items:
            for item in items:
                self.append(item)

    @classmethod
    def as_child(cls, parent, tag, item_type):
        '''
        Alternative callable constructor that instead defines this as a child object
        :param parent: Parent instance.
        :param tag: Tag of the child object
        :param item_type: Type of the object
        '''
        def helper(_ = None):
            return cls(parent.id + ':' + tag, item_type)
        return helper

    def __getitem__(self, index):
        '''
        Load an item by index where index is either an int or a slice
        No step greater than 1 is allowed for a slice.
        Warning: this is O(n))
        :param index: Number or a slice
        :return: The value if exists otherwise None
        '''
        if isinstance(index, slice):
            if index.step is not None and index.step != 1:
                raise NotImplementedError('Cannot specify a step to a RedisObject slice')
            return [
                self.decode_value(self.item_type, el)
                for el in self.redis.lrange(self.id, index.start, index.stop)
            ]
        else:
            return self.decode_value(self.item_type, self.redis.lindex(self.id, index))

    def __setitem__(self, index, val):
        '''
        Update an item by index
        Warning: this is O(n)
        :param index: Number
        :param val: New value
        :return: None
        '''
        self.redis.lset(self.id, index, self.encode_value(val))

    def __len__(self):
        '''
        :return: length of the list
        '''
        return self.redis.llen(self.id)

    def __delitem__(self, index):
        '''
        Delete an item from the list.
        (warning: this is O(n))
        :param index: Target
        :return: None
        '''
        self.redis.lset(self.id, index, '__DELETED__')
        self.redis.lrem(self.id, 1, '__DELETED__')

    def __iter__(self):
        '''
        Iterate over all items in this list.
        '''
        for el in self.redis.lrange(self.id, 0, -1):
            yield self.decode_value(self.item_type, el)

    def lpop(self):
        '''
        Remove and return the first item of the list
        '''
        return self.decode_value(self.item_type, self.redis.lpop(self.id))

    def rpop(self):
        '''
        Remove and return the last item of the list
        :return:
        '''
        return self.decode_value(self.item_type, self.redis.rpop(self.id))

    def lpush(self, val):
        '''
        Push val onto the head of the list
        :param val:
        :return: None
        '''
        self.redis.lpush(self.id, self.encode_value(val))

    def rpush(self, val):
        '''
        Push val onto the tail of the list
        :param val:
        :return: None
        '''
        self.redis.rpush(self.id, self.encode_value(val))

    def append(self, val):
        '''
        An alias to rpush
        :param val:
        :return:
        '''
        self.rpush(val)

    def insert_before(self, ref, val):
        '''
        Insert val in list name immediately before ref
        :param ref:
        :param val:
        :return: the new length of the list on success or -1 if ref is not in the list.
        '''
        return self.redis.linsert(self.id, 'BEFORE', ref, val)

    def insert_after(self, ref, val):
        '''
        Insert val in list name immediately after ref
        :param ref:
        :param val:
        :return: the new length of the list on success or -1 if ref is not in the list.
        '''
        return self.redis.linsert(self.id, 'AFTER', ref, val)

    def trim(self, start, stop):
        '''
        Trim the list by removing all values not within the slice between start and end
        start and end can be negative numbers just like Python slicing notation
        :param start:
        :param stop:
        :return: None
        '''
        self.redis.ltrim(self.id, start, stop)



from setuptools import setup

setup(
    name='pydis',
    packages=['pydis'],
    version='0.3',
    description='Python wrappers for StrictRedis',
    author='masayang',
    author_email='masayang@msushi.com',
    url='https://github.com/masayang/py_redis',
    install_requires=['redis', 'python_dotenv'],
    zip_safe=False
)
